"""
Django settings for cri_intranet project.

Generated by 'django-admin startproject' using Django 2.1.7.

For more information on this file, see
https://docs.djangoproject.com/en/2.1/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/2.1/ref/settings/
"""

import os
import env
import sys

from django.contrib import messages
from django.utils.functional import lazy
from django.urls import reverse

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))


# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/2.1/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = env.get_secret("DJANGO_SECRET_KEY")

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = env.get_bool("DJANGO_DEBUG", False)

PROBES_IPS = env.get_list("DJANGO_PROBES_IP", ["0.0.0.0/0"])
ALLOWED_HOSTS = env.get_list("DJANGO_ALLOWED_HOSTS", [])
DEFAULT_DOMAIN = ALLOWED_HOSTS[0] if ALLOWED_HOSTS else "localhost"

DEFAULT_ADMIN_USERNAME = env.get_string("DJANGO_DEFAULT_ADMIN_USERNAME", "admin")
DEFAULT_ADMIN_PASSWORD = env.get_secret("DJANGO_DEFAULT_ADMIN_PASSWORD")
DEFAULT_ADMIN_UID = 1

# A list of all the people who get code error notifications.
ADMINS = [(mail, mail) for mail in env.get_list("DJANGO_ADMINS", [])]

# Redirect plain HTTP requests to HTTPS.
SECURE_SSL_REDIRECT = not DEBUG

# Set this to True to avoid transmitting the CSRF cookie over HTTP accidentally.
CSRF_COOKIE_SECURE = not DEBUG

# Set this to True to avoid transmitting the session cookie over HTTP accidentally.
SESSION_COOKIE_SECURE = not DEBUG

# See https://docs.djangoproject.com/en/3.0/ref/middleware/#referrer-policy
SECURE_REFERRER_POLICY = "strict-origin-when-cross-origin"

SECURE_PROXY_SSL_HEADER = ("HTTP_X_FORWARDED_PROTO", "https")

DATA_UPLOAD_MAX_NUMBER_FIELDS = None

TESTING = "test" in sys.argv

# Kerberos settings

SPNEGO_HOSTNAME = env.get_string("SPNEGO_HOSTNAME")
KERBEROS_ADMIN_KEYTAB = env.get_string("KERBEROS_ADMIN_KEYTAB")
KERBEROS_DEFAULT_REALM = env.get_string("KERBEROS_DEFAULT_REALM")
KERBEROS_REALMS = {}

for realm in set(env.get_list("KERBEROS_REALMS", [])) | set([KERBEROS_DEFAULT_REALM]):
    realm_varname = realm.replace(".", "-")
    KERBEROS_REALMS[realm] = {
        "ADMIN_PRINCIPAL": env.get_string(
            f"KERBEROS_{realm_varname}_ADMIN_PRINCIPAL",
            f"intranet/admin@{realm}",
        ),
        "DEFAULT_POLICY": env.get_string(
            f"KERBEROS_{realm_varname}_DEFAULT_POLICY", None
        ),
        "NEWPRINC_ON_USER_CREATION": env.get_bool(
            f"KERBEROS_{realm_varname}_NEWPRINC_ON_USER_CREATION",
            realm == KERBEROS_DEFAULT_REALM,
        ),
    }


# Moodle settings

MOODLE_SERVERS = {}

for moodle in set(env.get_list("MOODLE_SERVERS", [])):
    moodle_varname = moodle.replace(".", "-")
    MOODLE_SERVERS[moodle] = {
        "API_URL": env.get_string(f"MOODLE_{moodle_varname}_API_URL"),
        "API_TOKEN": env.get_string(f"MOODLE_{moodle_varname}_API_TOKEN"),
        "IDNUMBER_PREFIX": env.get_string(
            f"MOODLE_{moodle_varname}_IDNUMBER_PREFIX", ""
        ),
    }


# Application definition

INSTALLED_APPS = [
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.humanize",
    "webpack_loader",
    "collectfast",  # Must stay before `django.contrib.staticfiles`.
    "django.contrib.staticfiles",
    "phonenumber_field",
    "social_django",
    "rest_framework",
    "rest_framework_serializer_field_permissions",
    "django_filters",
    "drf_yasg",
    "import_export",
    "django_q",
    "crispy_forms",
    "crispy_bootstrap4",
    "ckeditor",
    "cri_auth",
    "cri_api",
    "cri_models",
    "cri_tasks",
    "cri_cards",
    "cri_frontend",
    "oidc_provider",  # Must stay after cri_auth
]

MIDDLEWARE = [
    "cri_intranet.middleware.XRealIPMiddleware",
    "cri_intranet.middleware.ProbesMiddleware",
]

if DEBUG and not TESTING:
    MIDDLEWARE += [
        "debug_toolbar.middleware.DebugToolbarMiddleware",
    ]

MIDDLEWARE += [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
    "cri_auth.middleware.CRIAuthExceptionMiddleware",
    "cri_auth.kerberos.spnego.SPNEGOMiddleware",
    "cri_auth.impersonate.middleware.ImpersonateMiddleware",
    "oidc_provider.middleware.SessionManagementMiddleware",
]

ROOT_URLCONF = "cri_intranet.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
                "cri_frontend.context_processors.search",
                "cri_frontend.context_processors.header_shortcuts",
                "cri_frontend.context_processors.custom_site_name",
            ]
        },
    }
]

WSGI_APPLICATION = "cri_intranet.wsgi.application"

STATICFILES_DIRS = (os.path.join(BASE_DIR, "assets"),)

WEBPACK_LOADER = {
    "DEFAULT": {
        "BUNDLE_DIR_NAME": "webpack_bundles/",
        "CACHE": not DEBUG,
        "STATS_FILE": os.path.join(BASE_DIR, "webpack-stats.json"),
        "POLL_INTERVAL": 0.1,
        "IGNORE": [r".+\.hot-update.js", r".+\.map"],
    }
}
if TESTING:
    WEBPACK_LOADER["DEFAULT"][
        "LOADER_CLASS"
    ] = "webpack_loader.loaders.FakeWebpackLoader"

LOG_LEVEL = env.get_string("DJANGO_LOG_LEVEL", "DEBUG" if DEBUG else "WARNING").upper()

# A data structure containing configuration information. The contents of this
# data structure will be passed as the argument to the configuration method
# described in LOGGING_CONFIG.

LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "django.server": {
            "()": "django.utils.log.ServerFormatter",
            "format": "[{server_time}] {message}",
            "style": "{",
        },
        "default": {
            "()": "logging.Formatter",
            "format": "[{asctime}][{levelname}:{name}] {message}",
            "style": "{",
        },
    },
    "filters": {
        "require_debug_false": {"()": "django.utils.log.RequireDebugFalse"},
        "not_in_testing_mode": {"()": "cri_intranet.tests.NotInATestFilter"},
    },
    "handlers": {
        "console": {
            "level": "DEBUG",
            "class": "logging.StreamHandler",
            "filters": ["not_in_testing_mode"],
            "formatter": "default",
        },
        "django.server": {
            "level": "DEBUG",
            "class": "logging.StreamHandler",
            "formatter": "django.server",
        },
        "mail_admins": {
            "level": "ERROR",
            "filters": ["require_debug_false"],
            "class": "django.utils.log.AdminEmailHandler",
        },
    },
    "loggers": {
        "django": {
            "handlers": ["console", "mail_admins"],
            "level": LOG_LEVEL,
            "propagate": False,
        },
        "django.server": {
            "handlers": ["django.server"],
            "level": "INFO",
            "propagate": False,
        },
    },
    "root": {"handlers": ["console"], "level": LOG_LEVEL},
}


# Database
# https://docs.djangoproject.com/en/2.1/ref/settings/#databases

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.postgresql",
        "NAME": env.get_string("DJANGO_DB_NAME", "cri_intranet"),
        "USER": env.get_string("DJANGO_DB_USER", "cri_intranet"),
        "PASSWORD": env.get_secret("DJANGO_DB_PASSWORD"),
        "HOST": env.get_string("DJANGO_DB_HOST"),
        "PORT": env.get_string("DJANGO_DB_PORT", "5432"),
    },
    "ldap": {
        "ENGINE": "ldapdb.backends.ldap",
        "NAME": env.get_string("DJANGO_LDAP_URI"),
        "USER": env.get_string("DJANGO_LDAP_USER"),
        "PASSWORD": env.get_secret("DJANGO_LDAP_PASSWORD"),
    },
}

DATABASE_ROUTERS = ["ldapdb.router.Router"]

DEFAULT_AUTO_FIELD = "django.db.models.AutoField"

LDAP_BASE_DN = env.get_string("DJANGO_LDAP_BASE_DN")


# Caches
# https://docs.djangoproject.com/en/3.0/topics/cache/

CACHES = {
    "default": {
        "BACKEND": "django_redis.cache.RedisCache",
        "LOCATION": env.get_string("DJANGO_CACHE_URL"),
        "OPTIONS": {"CLIENT_CLASS": "django_redis.client.DefaultClient"},
    }
}


# Password validation
# https://docs.djangoproject.com/en/2.1/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        "NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator"  # noqa: E501
    },
    {
        "NAME": "django.contrib.auth.password_validation.MinimumLengthValidator",
        "OPTIONS": {"min_length": 10},
    },
    {
        "NAME": "cri_auth.validators.CharacterClassesValidator",
        "OPTIONS": {"min_classes": 3},
    },
    {"NAME": "django.contrib.auth.password_validation.CommonPasswordValidator"},
    {"NAME": "django.contrib.auth.password_validation.NumericPasswordValidator"},
]


# Internationalization
# https://docs.djangoproject.com/en/2.1/topics/i18n/

LANGUAGE_CODE = "en-us"

TIME_ZONE = "Europe/Paris"

USE_I18N = True

USE_L10N = False

USE_TZ = True

DATETIME_FORMAT = "Y-m-d H:i:s"

DATE_FORMAT = "Y-m-d"


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/2.1/howto/static-files/

STATIC_URL = "http://127.0.0.1:9000/cri-intranet/"

COLLECTFAST_DEBUG = DEBUG
COLLECTFAST_STRATEGY = "collectfast.strategies.boto3.Boto3Strategy"
STATICFILES_STORAGE = "cri_intranet.custom_storage.PublicS3Storage"
DEFAULT_FILE_STORAGE = "cri_intranet.custom_storage.PublicS3Storage"

AWS_ACCESS_KEY_ID = env.get_secret("S3_ACCESS_KEY")
AWS_SECRET_ACCESS_KEY = env.get_secret("S3_SECRET_KEY")
AWS_STORAGE_BUCKET_NAME = env.get_string("S3_BUCKET", "cri-intranet")
AWS_S3_ENDPOINT_URL = env.get_string("S3_ENDPOINT")
AWS_S3_URL_PROTOCOL = "https:" if env.get_bool("S3_SECURE_URLS", True) else "http:"
# S3 adressing style : auto, virtual or path
AWS_S3_ADDRESSING_STYLE = env.get_bool("S3_ADDRESSING_STYLE", "path")

AWS_QUERYSTRING_AUTH = True

AWS_STORAGE_PHOTOS_BUCKET_NAME = env.get_string(
    "S3_PHOTOS_BUCKET", "cri-intranet-photos"
)

# Settings only applied to public buckets
AWS_PUBLIC_CUSTOM_DOMAIN = env.get_string("S3_CUSTOM_DOMAIN", "") or None
AWS_PUBLIC_OBJECT_PARAMETERS = {"ACL": "public-read"}


# Tests

TEST_RUNNER = "cri_intranet.tests.CRIIntranetTestRunner"
TEST_OUTPUT_DIR = os.path.join(BASE_DIR, "test-reports/")


MESSAGE_TAGS = {
    messages.DEBUG: "secondary",
    messages.INFO: "info",
    messages.SUCCESS: "success",
    messages.WARNING: "warning",
    messages.ERROR: "danger",
}


# Django Auth

AUTH_USER_MODEL = "cri_models.CRIUser"
LOGIN_URL = "login"

LOGIN_REDIRECT_URL = "criuser_self"
LOGOUT_REDIRECT_URL = lazy(lambda url: reverse(url) + "?spnego=0")("login")


# Django debug toolbar


# Do not replace this function with a lambda, it breaks the tests subsystem
# which relies on its capability to override settings such as DEBUG.
def show_toolbar(request):
    from django.conf import settings  # pylint:disable=import-outside-toplevel

    return settings.DEBUG


DEBUG_TOOLBAR_CONFIG = {
    "SHOW_TOOLBAR_CALLBACK": "cri_intranet.settings.show_toolbar",
    "SHOW_COLLAPSED": True,
    "DISABLE_PANELS": {"debug_toolbar.panels.redirects.RedirectsPanel"},
}

DEBUG_TOOLBAR_PANELS = (
    "debug_toolbar.panels.versions.VersionsPanel",
    "debug_toolbar.panels.timer.TimerPanel",
    "debug_toolbar.panels.settings.SettingsPanel",
    "debug_toolbar.panels.headers.HeadersPanel",
    "debug_toolbar.panels.request.RequestPanel",
    "debug_toolbar.panels.sql.SQLPanel",
    "debug_toolbar.panels.staticfiles.StaticFilesPanel",
    "debug_toolbar.panels.templates.TemplatesPanel",
    "debug_toolbar.panels.cache.CachePanel",
    "debug_toolbar.panels.signals.SignalsPanel",
    "debug_toolbar.panels.logging.LoggingPanel",
    "debug_toolbar.panels.redirects.RedirectsPanel",
    "debug_toolbar.panels.profiling.ProfilingPanel",
    "mail_panel.panels.MailToolbarPanel",
)

# Email settings

DEFAULT_FROM_EMAIL = env.get_string(
    "DJANGO_DEFAULT_FROM_EMAIL", f"noreply@{DEFAULT_DOMAIN}"
)

if DEBUG and not TESTING:
    INSTALLED_APPS += [
        "debug_toolbar",
        "mail_panel",
    ]
    EMAIL_BACKEND = "mail_panel.backend.MailToolbarBackend"
else:
    EMAIL_HOST = env.get_string("DJANGO_SMTP_HOSTNAME", "localhost")
    EMAIL_PORT = env.get_string("DJANGO_SMTP_PORT", 25)
    EMAIL_HOST_USER = env.get_string("DJANGO_SMTP_USER", "")
    EMAIL_HOST_PASSWORD = env.get_secret("DJANGO_SMTP_PASSWORD", "")
    EMAIL_USE_TLS = env.get_bool("DJANGO_SMTP_STARTTLS", False)

SERVER_EMAIL = env.get_string("DJANGO_SERVER_EMAIL", f"intranet@{DEFAULT_DOMAIN}")
EMAIL_SUBJECT_PREFIX = env.get_string("DJANGO_EMAIL_SUBJECT_PREFIX", "[Django] ")


# OIDC Provider

OIDC_IDTOKEN_SUB_GENERATOR = "cri_auth.oidc_provider.criuser_sub_generator"
OIDC_AFTER_USERLOGIN_HOOK = "cri_auth.oidc_provider.after_userlogin_hook"
OIDC_EXTRA_SCOPE_CLAIMS = "cri_auth.oidc_provider.CRIScopeClaims"
OIDC_SESSION_MANAGEMENT_ENABLE = True
OIDC_IDTOKEN_INCLUDE_CLAIMS = True
OIDC_TEMPLATES = {"authorize": "cri_auth/oidc/authorize.html"}

# DRF

REST_FRAMEWORK = {
    "DEFAULT_AUTHENTICATION_CLASSES": [
        "rest_framework.authentication.SessionAuthentication",
        "rest_framework.authentication.BasicAuthentication",
        "cri_auth.kerberos.spnego.drf.SPNEGOAuthentication",
        "cri_auth.oidc_provider.drf.OIDCAccessTokenAuthentication",
    ],
    "DEFAULT_PERMISSION_CLASSES": ["rest_framework.permissions.DjangoModelPermissions"],
    "DEFAULT_FILTER_BACKENDS": ["django_filters.rest_framework.DjangoFilterBackend"],
    "DEFAULT_THROTTLE_CLASSES": [
        "rest_framework.throttling.AnonRateThrottle",
        "cri_api.throttle.UserRateThrottle",
    ],
    "DEFAULT_THROTTLE_RATES": {"anon": "100/hour", "user": "1000/hour"},
    "DEFAULT_PAGINATION_CLASS": "rest_framework.pagination.LimitOffsetPagination",
    "PAGE_SIZE": 100,
    "TEST_REQUEST_DEFAULT_FORMAT": "json",
}

SWAGGER_SETTINGS = {
    "REFETCH_SCHEMA_WITH_AUTH": True,
    "REFETCH_SCHEMA_ON_LOGOUT": True,
    "DEEP_LINKING": True,
    "LOGOUT_URL": "logout",
    "SECURITY_DEFINITIONS": {
        "Basic": {"type": "basic"},
        "Bearer": {"type": "apiKey", "name": "Authorization", "in": "header"},
    },
}

CRISPY_TEMPLATE_PACK = "bootstrap4"

SOCIAL_AUTH_PIPELINE = (
    "social_core.pipeline.social_auth.social_details",
    "social_core.pipeline.social_auth.social_uid",
    "social_core.pipeline.social_auth.auth_allowed",
    "cri_auth.pipeline.associate_by_email",
    "cri_auth.pipeline.protect_staff_accounts",
    "cri_auth.pipeline.forbid_service_accounts",
    "cri_auth.pipeline.force_password_login_on_unsynced_principal",
    "social_core.pipeline.social_auth.social_user",
    "social_core.pipeline.social_auth.associate_user",
)

SOCIAL_AUTH_LOGIN_ERROR_URL = LOGIN_URL

SOCIAL_AUTH_AZUREAD_OAUTH2_RESOURCE = "https://graph.microsoft.com/"

AUTHENTICATION_BACKENDS = (
    "cri_auth.backends.CRIModelBackend",
    "cri_auth.ldap.CRILDAPBackendWithPasswordUpdate",
    "cri_auth.kerberos.spnego.SPNEGOBackend",
    "cri_auth.backends.CRIAzureADOauth2",
    "cri_auth.impersonate.backend.ImpersonateBackend",
)

PASSWORD_HASHERS = [
    "django.contrib.auth.hashers.Argon2PasswordHasher",
    "django.contrib.auth.hashers.PBKDF2PasswordHasher",
    "django.contrib.auth.hashers.PBKDF2SHA1PasswordHasher",
    "django.contrib.auth.hashers.BCryptSHA256PasswordHasher",
]

SOCIAL_AUTH_AZUREAD_OAUTH2_KEY = env.get_string("AZUREAD_OAUTH2_KEY")
SOCIAL_AUTH_AZUREAD_OAUTH2_SECRET = env.get_secret("AZUREAD_OAUTH2_SECRET")
SOCIAL_AUTH_AZUREAD_OAUTH2_WHITELISTED_DOMAINS = env.get_list("AZUREAD_DOMAINS")
SOCIAL_AUTH_AZUREAD_OAUTH2_FIELDS_STORED_IN_SESSION = ("bypass_principal_check",)

AUTH_LDAP_SERVER_URI = env.get_string("DJANGO_LDAP_URI")
AUTH_LDAP_NO_NEW_USERS = True
AUTH_LDAP_ALWAYS_UPDATE_USER = False

MEILISEARCH = {
    "ADMIN_API_KEY": env.get_secret("MEILI_ADMIN_API_KEY", ""),
    "PUBLIC_API_KEY": env.get_secret("MEILI_PUBLIC_API_KEY", ""),
    "SERVER_URI": env.get_string("MEILI_SERVER_URI"),
    "PUBLIC_SERVER_URI": env.get_string("MEILI_PUBLIC_SERVER_URI"),
}

# phonenumber_field

PHONENUMBER_DEFAULT_REGION = "FR"

# import_export

IMPORT_EXPORT_USE_TRANSACTIONS = True
IMPORT_EXPORT_IMPORT_PERMISSION_CODE = "change"
IMPORT_EXPORT_EXPORT_PERMISSION_CODE = "view"

# Django Q

Q_CLUSTER = {
    "name": env.get_string("DJANGO_Q_CLUSTER_NAME", "cri_intranet"),
    "workers": env.get_int("DJANGO_Q_THREADS", 2),
    "timeout": env.get_int("DJANGO_Q_TIMEOUT", 600),
    "retry": env.get_int("DJANGO_Q_RETRY", 660),
    "django_redis": "default",
}

# ckeditor

CKEDITOR_CONFIGS = {"default": {"width": "100%"}}

NOTES_EXTRA_LINKS = [
    {
        "title": "Last activity",
        "type": "date",
        "icon": "far fa-clock",
        "content": "last_activity",
    },
    {
        "title": "Tickets from {username}",
        "icon": "far fa-check-square",
        "type": "link",
        "content": "https://tickets.cri.epita.fr/Search/Results.html?Class=RT%3A%3ATickets&Query=Requestor.EmailAddress+%3D+%27{email}%27",  # noqa: E501
        "external": True,
    },
]

PHOTO_THUMB_WIDTH = 500
PHOTO_SQUARE_SIZE = 500

CONNECTION_LOG_ENTRY_MAX_AGE = 365

SITE_NAME = env.get_string("SITE_NAME", "CRI intranet")
CRICARD_EXPORT_DENSITY = env.get_int("CRICARD_EXPORT_DENSITY", 300)
